!  liloload: simple loader which loads LILO first stage loader
!  Copyright (C) 1997-2000  Ferdy Hanssen <hanssen@croky.net>
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!
! version 1.8 - searches for a LILO bootloader and starts it

#define TOS_SEG		0x0700
#define TOS_OFF		0x0c00
#define CODESEG		0x0800
#define BOOTSEG		0x07c0
#define PARTSEG		0x0900
#define EXTPARTSEG	0x0920
#define PART1OFF	0x01be
#define PART2OFF	0x01ce
#define PART3OFF	0x01de
#define PART4OFF	0x01ee
#define BIOSSEG		0x0040
#define NUMOFHDS	0x0075

! 0x07000-0x07BFF   3.0 kB   stack
! 0x07C00-0x07DFF   0.5 kB   boot load address
! 0x07E00-0x07E08   0.0 kB   program variables (before move)
! 0x08000-0x081FF   0.5 kB   this program (after move)
! 0x08200-0x08208   0.0 kB   program variables (after move)
! 0x08400-0x085FF   0.5 kB   floppy debug extra program space
! 0x09000-0x091FF   0.5 kB   MBR (with partition table)
! 0x09200-0x093FF   0.5 kB   extended partition blocks

! Partition table: 4 entries of 16 bytes, starting at offset 0x01be
! from the beginning of the sector.  Each entry is layed out as
! follows:
!
! Offset    Type     Function
!  0        byte     Boot indicator: 0x80 = active, 0x00 = inactive
!  1        byte     Starting head
!  2        byte     Starting sector
!  3        byte     Starting cylinder
!  4        byte     Partition type
!  5        byte     Ending head
!  6        byte     Ending sector
!  7        byte     Ending cylinder
!  8        long     Starting sector counting from 0
! 12        long     Number of sectors in partition

#ifdef FLOPPYDEBUG
# ifndef DEBUG
#  define DEBUG
# endif
#endif

	.text

	.globl	_main

	.org	0

_main:	cli				! No interrupts
	mov	ax,#TOS_SEG
	mov	ss,ax
	mov	sp,#TOS_OFF		! Set stack pointer
	sti				! Interrupts safe again

	mov	ax,#BOOTSEG		! Set DS
	mov	ds,ax

	mov	ext_es,es		! Copy possible external parameters
	mov	ext_si,si
	mov	ext_bx,bx
	mov	ext_dl,dl

	mov	ax,#CODESEG		! Beam us up...
	mov	es,ax
	mov	cx,#260			! Program space (256 words) +
					! variables (7 bytes)
	xor	si,si
	xor	di,di
	cld
	rep
	movsw
	jmpi	go,CODESEG

go:	mov	ds,ax			! AX is already set

#ifdef FLOPPYDEBUG
	call	read_rest		! Read the rest of the program
#else
	mov	si,#msg_welcome		! Welcome text
	call	printstr
#endif

	mov	ax,#BIOSSEG
	mov	es,ax
	seg	es
	mov	al,NUMOFHDS		! Get number of HDs present
	xor	ah,ah
	add	ax,#0x7f
	mov	maxdrive,ax		! Transform into highest drive possible

	mov	si,#0x80		! Set SI:7-0 to first HD, SI:7-0 is
					! drive loop counter, points to drive
					! we investigate

driveloop:
#ifdef DEBUG
	call	debug_show_drive
#endif
	mov	ax,#PARTSEG		! Read partition table into memory
					! at PARTSEG:0
	mov	cx,#1			! track 0, sector 1
	mov	dx,si			! head 0, drive SI:7-0
	call	read1sector
	seg	es
	cmp	0x1fe,#0xaa55		! Is it a valid partition table?
#ifndef DEBUG
	jne	next_drive		! If not, next drive
#else
	je	mbrloaded
	mov	ax,#DBGmbrnotloaded
	call	DBGprintstr
	jmp	next_drive
mbrloaded:
	mov	ax,#DBGmbrloaded
	call	DBGprintstr
#endif

	xor	bp,bp			! BP is part. loop counter, contains
					! partition number we investigate
partloop:
#ifdef DEBUG
	call	debug_show_part
#endif
	mov	di,bp
	shl	di,#4
	add	di,#PART1OFF		! DI points to partition data
	mov	ax,#PARTSEG
	mov	es,ax
#ifndef FLOPPYDEBUG
	seg	es
	mov	ah,(di+1)		! Get cylinder
	seg	es
	mov	al,(di+3)
	shr	ah,#6
	cmp	ax,#1023		! Is it a valid cylinder (1023 indicates
					! cylnder >= 1023)
	je	next_prim_partition	! If not, skip it
#else
	br	getcylinder1
approvedcylinder1:
#endif
	seg	es
	mov	al,(di+4)		! Get partition type
	cmp	al,#0x83		! Is it of type "Linux native"?
	jne	check_ext_partition	! If no, check for extended partition

	call	try_for_lilo		! Try for LILO

next_prim_partition:
	inc	bp
	cmp	bp,#4			! Have we had all four partitions?
	jl	partloop		! If not, try next one
					! Otherwise, try next drive

next_drive:
	inc	si
	cmp	si,maxdrive		! Have we had all drives?
	jle	driveloop		! If not, try next one
					! Otherwise, not found

! Print "not found" and halt.
	mov	si,#msg_notfound
	jmp	error

check_ext_partition:
! According to Linux kernel source (linux/include/linux/genhd.h), 0x05, 0x0f,
! and 0x85 can denote an extended partition
	cmp	al,#0x0f		! Is it of type "Extended 0x0f"?
#ifndef DEBUG
	je	try_ext_partition	! If yes, try for LILO
#else
	je	try_ext_dbg
#endif
	and	al,#0x7f
	cmp	al,#0x05		! Is it of type "Extended 0x05 or 0x85"?
#ifndef DEBUG
	jne	next_prim_partition	! If not, try next primary partition
#else
	je	try_ext_dbg
	mov	ax,#DBGtypenotok
	call	DBGprintstr
	jmp	next_prim_partition
try_ext_dbg:
	mov	byte DBGpartnum,#0x41	! Extended partitions count from A
#endif
try_ext_partition:
	call	try_for_lilo
ext_partition_loop:
#ifdef DEBUG
	call	debug_show_ext_part
	inc	byte DBGpartnum
#endif
	mov	dx,si			! (Drive)
	seg	es			! Read first sector of extended
	mov	cx,(di+2)		! partition
	seg	es
	mov	dh,(di+1)
	mov	ax,#EXTPARTSEG
	call	read1sector
	mov	di,#PART1OFF		! First partition entry contains
					! logical partition
#ifndef FLOPPYDEBUG
	seg	es
	mov	ah,(di+1)		! Get cylinder
	seg	es
	mov	al,(di+3)
	shr	ah,#6
	cmp	ax,#1023		! Is it a valid cylinder (1023 indicates
					! cylnder >= 1023)
	je	next_log_partition	! If not, skip it
#else
	br	getcylinder2
approvedcylinder2:
#endif
	seg	es
	cmp	byte (di+4),#0x83	! Is it of type "Linux native"?
#ifndef DEBUG
	jne	next_log_partition	! If not, try next logical partition
#else
	je	try_log_lilo_dbg
	mov	ax,#DBGtypenotok
	call	DBGprintstr
	jmp	next_log_partition
try_log_lilo_dbg:
#endif
	call	try_for_lilo		! Try for LILO
next_log_partition:
	mov	di,#PART2OFF		! Second partition entry may contain
					! pointer to next logical partition
	seg	es
	cmp	byte (di+4),#0x00	! Last logical partition in chain?
	je	next_prim_partition	! If yes, try next primary partition
	jmp	ext_partition_loop	! Try next logical partition

! Print "read error" and halt.
read_error:
	mov	si,#msg_readerror

error:
	call	printstr

! Stop here, and reboot when ESC is pressed.
halt:
	mov	si,#msg_esc		! Print informatory message
	call	printstr
waitforesc:
	xor	ah,ah			! Wait for key
	int	0x16
#ifndef DEBUG
	cmp	al,#0x1b		! ESC pressed?
	jne	waitforesc		! If not, wait
#endif
	int	0x19			! Reload and re-execute boot sector

! Read 1 sector from disk at AX:0, returns on success.
! Needs cylinder in CH, head in DH, sector in CL, drive in DL.
! Leaves AX in ES.
read1sector:
	xor	bx,bx			! Result at ax:0
	mov	es,ax
#ifdef FLOPPYDEBUG
	call	DBGprint_sector
#endif
	mov	ax,#0x0201		! Read one sector
	int	0x13
	jc	read_error
	ret

! Check if bootsector of partition contains
! LILO, if so, start it, otherwise, return to caller
! Needs partition entry in ES:DI.
! Does not destroy ES, DI, SI.
try_for_lilo:
	push	es			! Do not destroy ES
#ifdef DEBUG
	mov	ax,#DBGtypeok
	call	DBGprintstr
#endif
	mov	dx,si			! Drive in SI:7-0
	seg	es			! Read first sector of partition
	mov	cx,(di+2)		! Cylinder + sector
	seg	es
	mov	dh,(di+1)		! Head
	mov	ax,#BOOTSEG		! Read at BOOTSEG:0
	call	read1sector

	seg	es
	cmp	0x1fe,#0xaa55		! Is it a valid bootsector?
	jne	no_lilo			! If not, return

lilo20:
	seg	es
	cmp	0x2,#0x494c		! Is it the LILO (<= 0.20) boot loader?
	jne	lilo21			! If not, try LILO (>= 0.21)
	seg	es
	cmp	0x4,#0x4f4c
	je	loaded			! If yes, start it.

lilo21:
	seg	es
	cmp	0x6,#0x494c		! Is it the LILO (>= 0.21) boot loader?
	jne	no_lilo			! If not, return
	seg	es
	cmp	0x8,#0x4f4c
	je	loaded			! If yes, start it.
no_lilo:
	pop	es
	ret

! LILO present at BOOTSEG:0
! Do not care about stack consistency any more, just start it.
loaded:
	mov	es,ext_es		! Restore external parameters
	mov	si,ext_si
	mov	bx,ext_bx
	mov	dl,ext_dl

	jmpi	0,BOOTSEG		! Start LILO

! Print string at (SI)
! Does not destroy ES, DI, BP.
printstr:
	push	es
	push	bp
	push	di
	xor	bh,bh
	mov	ah,#0x0e
printstr0:
	lodsb
	or	al,al
	jz	printstr1
	push	si
	int	0x10
	pop	si
	jmp	printstr0
printstr1:
	pop	di
	pop	bp
	pop	es
	ret

#ifdef DEBUG
! Prints P#<partnum>, partnum is taken from BP.
! Does not destroy SI, BP.
debug_show_part:
	mov	ax,bp
	add	al,#0x31
	mov	DBGpartnum,al

debug_show_ext_part:
	mov	ax,#DBGtrypart

! Print string at (AX)
! Does not destroy ES, DI, SI, BP.
DBGprintstr:
	push	si
	mov	si,ax
	call	printstr
	pop	si
	ret
#endif

msg_readerror:
#ifndef DEBUG
	.ascii	"Read error"
#else
	.ascii	"RE"
#endif
	.byte	0
msg_notfound:
#ifndef DEBUG
	.ascii	"Not found"
#else
	.ascii	"NF"
#endif
	.byte	0
msg_esc:
#ifndef DEBUG
	.ascii	", press ESC to reboot"
#else
	.ascii	"!"
msg_welcome:
	.byte	13,10,0
#endif
	.byte	0
#ifndef DEBUG
msg_welcome:
	.byte	13,10
	.ascii	"Loading LILO"
	.byte	13,10,0
#endif

#ifdef DEBUG
DBGtrydrive:
	.ascii	"D#"
DBGdrivenum:
	.ascii	"00"
	.byte	0
DBGmbrloaded:
	.ascii	", p"
DBGmbrnotloaded:
	.ascii	"; "
	.byte	0
DBGtrypart:
	.ascii	"P#"
DBGpartnum:
	.ascii	"0"
	.byte	0
DBGtypeok:
	.ascii	", OK"
DBGtypenotok:
	.ascii	"; "
	.byte	0
#endif

#ifndef DEBUG
	.ascii	"LILOloader V1.8; "
	.ascii	"(C) 1997-2000 by "
	.ascii	"Ferdy Hanssen <hanssen@croky.net>"
	.byte	0
#endif

#ifdef DEBUG
! Returns ascii representation of hex AL:3-0 in AL.
altohex:
	cmp	al,#0x0a
	jl	altohex0
	add	al,#0x27
altohex0:
	add	al,#0x30
	ret

! Prints D#<drivenum>, drivenum is taken from SI.
! Does not destroy SI, BP.
debug_show_drive:
	mov	ax,si
	and	al,#0x0f
	call	altohex
	mov	DBGdrivenum+1,al
	mov	ax,si
	and	al,#0xf0
	shr	al,#4
	call	altohex
	mov	DBGdrivenum,al
	mov	ax,#DBGtrydrive
	jmp	DBGprintstr
#endif

#ifdef FLOPPYDEBUG
! Read the rest of the floppy debug program
read_rest:
	mov	printflpmsg,#0xc3	! Put a ret instruction there
	mov	ax,#CODESEG+0x40	! Read second sector into memory at
					! CODESEG+0x40:0
	mov	es,ax
	xor	bx,bx
	mov	cx,#3			! track 0, sector 3
	mov	dx,#0			! head 0, drive 0 (first floppy)
	mov	ax,#0x0201		! Read one sector from
	int	0x13
	br	printflpmsg
#endif

	.org	0x1fe

	.word	0xaa55			! boot sector magic

! The following bytes will not be stored, they are variables

! The following four variables are stored before the move,
! so they have to be copied as well.

ext_si:	.word	0			! External interface
ext_es:	.word	0
ext_bx:	.word	0
ext_dl:	.byte	0

! The following variables are stored after the move,
! so they do not have to be copied.

maxdrive:
	.word	0x80			! Last drive to check

#ifdef FLOPPYDEBUG
	.org	0x400			! Set address right

! Set screen to 80x50 and print welcome message
printflpmsg:
	mov	ax,#0x0003
	int	0x10
	mov     ax,#0x1112
	xor     bl,bl
	int     0x10
	mov     ax,#0x1200
	mov     bl,#0x20
	int     0x10
	mov     ax,#0x1201
	mov     bl,#0x34
	int     0x10
	mov     ah,#0x01
	mov     cx,#0x0607
	int     0x10
	mov	ax,#DBGflp
	br	DBGprintstr

DBGflp:
	.ascii	"Floppy debug version"
	.byte	13,10,0

getcylinder1:
	seg	es
	mov	ah,(di+1)		! Get cylinder
	seg	es
	mov	al,(di+3)
	shr	ah,#6
	cmp	ax,#1023		! Is it a valid cylinder (1023 indicates
					! cylnder >= 1023)
	je	skipcylinder1		! If not, skip it
	br	approvedcylinder1
skipcylinder1:
	mov	ax,#DBGskippart
	call	DBGprintstr
	br	next_prim_partition

getcylinder2:
	seg	es
	mov	ah,(di+1)		! Get cylinder
	seg	es
	mov	al,(di+3)
	shr	ah,#6
	cmp	ax,#1023		! Is it a valid cylinder (1023 indicates
					! cylnder >= 1023)
	je	skipcylinder2		! If not, skip it
	br	approvedcylinder2
skipcylinder2:
	mov	ax,#DBGskippart
	call	DBGprintstr
	br	next_log_partition

DBGskippart:
	.ascii	", SKIP"
	.byte	0

! Put ascii equivalent of hex ax on (bx)
DBGprint_ax_on_bx:
	call	DBGprint_ah_on_bx
	add	bx,#2

! Put ascii equivalent of hex al on (bx)
DBGprint_al_on_bx:
	push	ax
	and	al,#0x0f
	call	altohex
	mov	(bx+1),al
	pop	ax
	push	ax
	and	al,#0xf0
	shr	al,#4
	call	altohex
	mov	(bx),al
	pop	ax
	ret

! Put ascii equivalent of hex ah on (bx)
DBGprint_ah_on_bx:
	push	ax
	mov	al,ah
	call	DBGprint_al_on_bx
	pop	ax
	ret

! Print values of ES, DI, and SI
DBGprint_es_di_si:
	pusha
	mov	bx,#DBGes_string
	mov	ax,es
	call	DBGprint_ax_on_bx
	mov	bx,#DBGdi_string
	mov	ax,di
	call	DBGprint_ax_on_bx
	mov	bx,#DBGsi_string
	mov	ax,si
	call	DBGprint_ax_on_bx
	mov	ax,#DBGes_di_si_string
	call	DBGprintstr
	popa
	ret

DBGes_di_si_string:
	.ascii	"(ES="
DBGes_string:
	.ascii	"0000"
	.ascii	",DI="
DBGdi_string:
	.ascii	"0000"
	.ascii	",SI="
DBGsi_string:
	.ascii	"0000)"
	.byte	0

! Print sector we are going to read from disk
DBGprint_sector:
	pusha
	mov	bx,#DBGdrive_string
	mov	ax,si
	call	DBGprint_al_on_bx
	mov	bx,#DBGcylinder_string
	mov	al,cl
	and	al,#0xc0
	shl	ax,#2
	mov	al,ch
	add	ax,#1
	call	DBGprint_ax_on_bx
	mov	bx,#DBGhead_string
	mov	al,dh
	call	DBGprint_al_on_bx
	mov	bx,#DBGsector_string
	mov	al,cl
	and	al,#0x3f
	call	DBGprint_al_on_bx
	mov	bx,#DBGsegment_string
	mov	ax,es
	call	DBGprint_ax_on_bx
	mov	ax,#DBGprint_sector_string
	call	DBGprintstr
	popa
	ret

DBGprint_sector_string:
	.ascii	"(D"
DBGdrive_string:
	.ascii	"00"
	.ascii	",C"
DBGcylinder_string:
	.ascii	"0000"
	.ascii	",H"
DBGhead_string:
	.ascii	"00"
	.ascii	",S"
DBGsector_string:
	.ascii	"00"
	.ascii	"->"
DBGsegment_string:
	.ascii	"0000"
	.ascii	":0000)"
	.byte	0

#endif
